from abc import ABC, abstractmethod
import functools
import json
from typing import Any, Callable
import asyncio
from bs_monitoring.common.utils import MonitoringServiceError, register_config
from dataclasses import dataclass, asdict


@dataclass
class AlertServiceConfig:
    """Configuration for the alert service component

    Args:
        type (str): The type of the alert service.
        config (Union[OpsgenieAlertServiceConfig, None], optional): The configuration for the alert service. Defaults to None.
    """

    type: str
    config: None = None


class AlertService(ABC):
    """Abstract class for alert services."""

    def __init__(self, config: Any) -> None:
        for k, v in asdict(config).items():
            setattr(self, k, v)

    @abstractmethod
    def send_alert(
        self,
        message: str,
        description: str | None = None,
    ):
        """Method to send an alert.

        Args:
            message (str): The message for the alert.
            description (Optional[str], optional): Extra description for the alert. Defaults to None.
        """
        pass


def alert(
    message: str,
) -> Callable:
    """Decorator to send an alert if an exception is raised.

    Args:
        message (str): The message to send in the alert.

    Returns:
        Callable: The decorated function.
    """

    def decorator(func: Callable):
        @functools.wraps(func)
        async def async_wrapper(self, *args, **kwargs):
            try:
                return await func(self, *args, **kwargs)
            except MonitoringServiceError as e:
                alert_service: AlertService = self.alert_service
                context = {
                    "class": self.__class__.__name__,
                    "method": func.__name__,
                    "error": str(e),
                }
                context = json.dumps(context)
                alert_service.send_alert(
                    message=message,
                    description=context,
                )
            except Exception as e:
                raise e

        @functools.wraps(func)
        def sync_wrapper(self, *args, **kwargs):
            try:
                return func(self, *args, **kwargs)
            except MonitoringServiceError as e:
                alert_service: AlertService = self.alert_service
                context = {
                    "class": self.__class__.__name__,
                    "method": func.__name__,
                    "error": str(e),
                }
                context = json.dumps(context)
                alert_service.send_alert(
                    message=message,
                    description=context,
                )
            except Exception as e:
                raise e

        if asyncio.iscoroutinefunction(func):
            return async_wrapper
        return sync_wrapper

    return decorator


def register_alert_service(cls):
    assert issubclass(
        cls, AlertService
    ), "AlertService must be a subclass of AlertService"

    name = (
        cls.__name__
        if not cls.__name__.endswith("AlertService")
        else cls.__name__.removesuffix("AlertService")
    )
    __ALERT_SERVICES[name] = cls
    register_config(cls, AlertServiceConfig)
    return cls


__ALERT_SERVICES = {}


def create_alert_service(config: AlertServiceConfig) -> AlertService:
    """Create an alert service based on the configuration.

    Args:
        config (AlertServiceConfig): The configuration for the alert service.

    Raises:
        Exception: If the alert service type is unknown.

    Returns:
        AlertService: The alert service subclass instance.
    """
    c = config.config if hasattr(config, "config") else None

    svc = __ALERT_SERVICES.get(config.type)
    if svc is not None:
        return svc(c)
    else:
        raise Exception(f"Unknown alert service type: {config.type}")



from typing import Any
import requests
from bs_monitoring.alert_services.base import AlertService, register_alert_service
from bs_monitoring.common.utils import ConfigField


@register_alert_service
class DiscordAlertService(AlertService):
    webhook_url = ConfigField(str)

    def __init__(self, config: Any):
        """Alert service that sends alerts to Discord via webhooks.
        
        To get the webhook URL:
        1. Go to Discord Server Settings
        2. Select "Integrations"
        3. Click on "Webhooks"
        4. Create a new webhook or use existing one
        5. Copy the Webhook URL

        Args:
            config: The configuration containing the webhook_url.
        """
        super().__init__(config)

    def build_alert(
        self,
        message: str,
        description: str | None = None,
        alias: str | None = None,
    ) -> dict[str, Any]:
        """Builds a Discord webhook message.

        Args:
            message (str): The message title for the alert.
            description (Optional[str]): Description for the alert. Defaults to None.
            alias (Optional[str]): Not used for Discord. Defaults to None.

        Returns:
            Dict[str, Any]: The Discord webhook message object.
        """
        embed = {
            "title": message,
            "description": description if description else message,
            "color": 15158332  # Red color for alerts
        }
        
        return {
            "embeds": [embed]
        }

    def send_alert(
        self,
        message: str,
        description: str | None = None,
    ):
        """Sends an alert to Discord via webhook.

        Args:
            message (str): Alert message
            description (Optional[str]): Alert description. Defaults to None.

        Raises:
            Exception: Raised if the alert could not be sent.
        """
        body = self.build_alert(message, description)
        
        response = requests.post(url=self.webhook_url, json=body)
        if response.status_code not in (200, 204):
            raise Exception(f"Failed to send alert to Discord: {response.text}") 



from datetime import datetime, timedelta
from elasticsearch import AsyncElasticsearch
from typing import Any
from dataclasses import dataclass
import pytz
from bs_monitoring.alert_services import alert, AlertService
from bs_monitoring.common.utils import ConfigField
from bs_monitoring.data_sources import DataSource, register_data_source


@dataclass
class ElasticDataSourceConfig:
    """Configuration for the Elasticsearch data source.

    Args:
        url (str): The URL for the Elasticsearch instance.
        indices (List[str]): The indices to consume messages from.
        basic_auth (Optional[List[str]], optional): Basic authentication credentials. Defaults to None.
        api_key (Optional[str], optional): API key for the Elasticsearch instance. Defaults to None.
        history_length (int, optional): The history length to consume messages from. Defaults to 1
    """

    url: str
    indices: list[str]
    basic_auth: list[str] | None = None
    api_key: str | None = None
    history_length: int = 1


@register_data_source
class ElasticDataSource(DataSource):
    basic_auth = ConfigField(list[str] | None, default=None)
    url = ConfigField(str)
    indices = ConfigField(list[str])
    api_key = ConfigField(str | None, default=None)
    history_length = ConfigField(int, default=1)

    def __init__(self, alert_service: AlertService, config: Any):
        """A data source that consumes messages from Elasticsearch.

        Args:
            args (ElasticDataSourceConfig): The configuration for the data source.
            alert_service (AlertService): The alert service to use.
        """
        super().__init__(config)
        auth = (
            {"basic_auth": tuple(self.basic_auth)}
            if self.basic_auth
            else {"api_key": self.api_key}
        )
        self.client_ = AsyncElasticsearch(self.url, **auth)
        self.indices = self.indices
        self.history_length_ = self.history_length
        self.alert_service = alert_service

        self.lookback = timedelta(days=self.history_length_)

    @alert(
        message="Failed to consume messages from Elasticsearch.",
    )
    async def produce(self) -> dict[str, list[dict[str, Any]]]:
        """Consumes messages from Elasticsearch since last check."""
        end = datetime.now(pytz.utc)
        start = end - self.lookback

        self.query_ = {
            "query": {
                "range": {
                    "@timestamp": {
                        "gte": int(start.timestamp() * 1000),
                        "lte": int(end.timestamp() * 1000),
                    }
                }
            },
            "sort": [{"@timestamp": {"order": "desc"}}],
        }

        events = {}
        for index in self.indices:
            all_hits = []
            resp = await self.client_.search(
                index=index, body=self.query_, scroll="5m", size=1000
            )

            scroll_id = resp["_scroll_id"]
            hits = resp["hits"]["hits"]
            all_hits.extend(hit["_source"] for hit in hits)

            while len(hits) > 0:
                resp = await self.client_.scroll(scroll_id=scroll_id, scroll="5m")
                scroll_id = resp["_scroll_id"]
                hits = resp["hits"]["hits"]
                all_hits.extend(hit["_source"] for hit in hits)

            events[index] = all_hits

        return events

    async def close(self) -> None:
        """Clean up Elasticsearch client connection."""
        await self.client_.close()



import yaml
import os
import re
from dataclasses import dataclass, field
from typing import Any
from dacite import from_dict, Config
from argparse import ArgumentParser, ArgumentTypeError

from bs_monitoring.alert_services import AlertServiceConfig
from bs_monitoring.monitors import MonitorConfig
from bs_monitoring.data_sources import DataSourceConfig
from bs_monitoring.databases import DatabaseConfig
# from bs_monitoring.common.utils import import_submodules


@dataclass
class RootConfig:
    """Configuration for the service, contains configurations for the different components.
    Is read from a YAML file.

    Args:
        DataSource (DataSourceConfig): The configuration for the data source component.
        AlertService (AlertServiceConfig): The configuration for the alert service component.
        Monitors (List[MonitorConfig]): The configuration for the monitor components.
    """

    DataSource: DataSourceConfig
    AlertService: AlertServiceConfig
    Monitors: list[MonitorConfig]
    Interval: int = 86400
    Databases: list[DatabaseConfig] = field(default_factory=list)


def read_yaml_config(file_path: str) -> RootConfig:
    """Reads a YAML configuration file and returns a RootConfig object.

    Args:
        file_path (str): The path to the YAML configuration file.

    Raises:
        ArgumentTypeError: Raised if the file cannot be read or the configuration is invalid.

    Returns:
        RootConfig: The configuration object.
    """

    def replace_env_vars(value: Any):
        if isinstance(value, str):
            return re.sub(
                r"\$\{([^}^{]+)\}",
                lambda x: os.environ.get(x.group(1), x.group(0)),
                value,
            )
        elif isinstance(value, dict):
            return {k: replace_env_vars(v) for k, v in value.items()}
        elif isinstance(value, list):
            return [replace_env_vars(v) for v in value]
        return value

    with open(file_path, "r") as file:
        try:
            dct = replace_env_vars(yaml.safe_load(file))

            config = from_dict(
                data_class=RootConfig, data=dct, config=Config(strict=True)
            )
            return config
        except yaml.YAMLError as e:
            raise ArgumentTypeError(f"Error parsing YAML file: {e}")
        except Exception as e:
            raise ArgumentTypeError(f"Error parsing configuration: {e}")


def read_config() -> RootConfig:
    """Parses the command line arguments and reads the configuration file.

    Returns:
        RootConfig: The configuration object.
    """
    parser = ArgumentParser(
        description="Service monitoring quality and quantity of messages from Kafka."
    )
    parser.add_argument(
        "--config",
        type=read_yaml_config,
        help="The path to the configuration file.",
        required=True,
    )
    return parser.parse_args().config



from typing import Any
from bs_monitoring.common.utils import MonitoringServiceError, ConfigField
from bs_monitoring.alert_services import alert, AlertService
from cerberus import Validator
import yaml

from bs_monitoring.monitors.base import Monitor, register_monitor


class DataSchemeError(MonitoringServiceError):
    def __init__(self, message: str) -> None:
        """Exception raised when there is an error in the data scheme.

        Args:
            message (str): The error message.
        """
        super().__init__(message)


@register_monitor
class DataSchemeMonitor(Monitor):
    file = ConfigField(str)

    def __init__(
        self,
        alert_service: AlertService,
        db_name: str | None = None,
        config: Any = None,
    ) -> None:
        """Monitor to check the data scheme of the data.

        Args:
            alert_service (AlertService): The alert service to use.
            config (DataSchemeMonitorConfig): The configuration for the monitor.
        """
        super().__init__(alert_service, db_name, config)

        with open(self.file, "r") as f:
            self.validator_ = Validator(yaml.safe_load(f), allow_unknown=True)

    @alert(message="Data scheme error")
    async def process(self, data: dict[str, Any]) -> None:
        """Method to process the data, sends an alert if the data does not match the data scheme.

        Args:
            data (Dict[str, Any]): The data to process.

        Raises:
            DataSchemeError: Raised if the data does not match the data scheme.
        """
        for k, v in data.items():
            invalid_items = list(filter(lambda x: not self.validator_.validate(x), v))
            if len(invalid_items) > 0:
                raise DataSchemeError(
                    f"Data scheme error {self.validator_.errors} for {k}, total invalid items: {len(invalid_items)}"
                )



import json
from typing import Any
import requests
from bs_monitoring.alert_services.base import AlertService, register_alert_service
from bs_monitoring.common.utils import ConfigField


@register_alert_service
class OpsgenieAlertService(AlertService):
    api_key = ConfigField(str)

    def __init__(self, config: Any):
        """Alert service that sends alerts to Opsgenie.

        Args:
            args (OpsgenieAlertServiceConfig): The configuration for the alert service.
        """
        super().__init__(config)

    def build_alert(
        self,
        message: str,
        description: str | None = None,
        alias: str | None = None,
    ) -> dict[str, Any]:
        """Builds an alert object.

        Args:
            message (str): The message for the alert.
            description (Optional[str], optional): Description for the alert. Defaults to None.
            alias (Optional[str], optional): Alias for the alert. Defaults to None.

        Returns:
            Dict[str, Any]: The alert object.
        """
        return {
            "message": message,
            "description": description,
            "alias": alias,
        }

    def send_alert(
        self,
        message: str,
        description: str | None = None,
    ):
        """Sends an alert to the alert service.

        Args:
            message (str): Alert message
            description (Optional[str], optional): Alert description. Defaults to None.
            alias (Optional[str], optional): Alert alias. Defaults to None.

        Raises:
            Exception: Raised if the alert could not be sent.
        """
        body = self.build_alert(message, description)
        body = {k: v for k, v in body.items() if v is not None}

        url = "https://api.opsgenie.com/v2/alerts"
        headers = {
            "Authorization": f"GenieKey {self.api_key}",
            "Content-Type": "application/json",
        }
        response = requests.post(url=url, headers=headers, data=json.dumps(body))
        if response.status_code != 202:
            raise Exception(f"Failed to send alert: {response.text}")



from typing import Any
from dataclasses import field, make_dataclass


class ConfigField:
    def __init__(self, field_type: type, default: Any = None):
        self.field_type = field_type
        self.default = default

    def _to_attribute(self) -> tuple[Any, Any] | Any:
        if self.default is not None:
            return (self.field_type, field(default=self.default))
        return (self.field_type,)

    def __set_name__(self, owner, name):
        self.name = name

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self
        return getattr(obj.config, self.name)

    def __set__(self, obj, value):
        setattr(obj.config, self.name, value)


def register_config(cls: type, config_type: type):
    config_fields = [
        (name, *value._to_attribute())
        for name, value in vars(cls).items()
        if isinstance(value, ConfigField)
    ]
    c = make_dataclass(cls.__name__ + "Config", config_fields)
    setattr(cls, "config", c)
    if not config_type.__annotations__[
        "config"
    ]:  # DON'T TOUCH, DACITE BREAKS UNLESS THIS IS HERE
        config_type.__annotations__["config"] = c
    else:
        config_type.__annotations__["config"] |= c


class MonitoringServiceError(Exception):
    """Base class for exceptions in the monitoring service."""

    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(message)




from abc import ABC, abstractmethod
from dataclasses import asdict, dataclass
from typing import Any, AsyncGenerator
import asyncio
from datetime import datetime

from bs_monitoring.alert_services import AlertService
from bs_monitoring.common.utils import register_config


@dataclass
class DataSourceConfig:
    """Configuration for the data source component

    Args:
        type (str): The type of the data source.
        config (Union[ElasticDataSourceConfig, None], optional): The configuration for the data source. Defaults to None.
    """

    type: str
    config: None = None


class DataSource(ABC):
    def __init__(self, config: Any) -> None:
        for k, v in asdict(config).items():
            setattr(self, k, v)
        self.last_run = None

    @abstractmethod
    async def produce(self) -> dict[str, list[dict[str, Any]]]:
        """Single production of data from the data source."""
        pass

    async def produce_continuous(self, interval_seconds: int = 86400) -> AsyncGenerator[dict[str, list[dict[str, Any]]], None]:
        """Continuously produce data at specified intervals using asyncio.
        
        Args:
            interval_seconds (float): Interval between productions in seconds. Defaults to 86400 (1 day).
        """
        while True:
            current_time = datetime.now()
            
            if self.last_run is None or (current_time - self.last_run).total_seconds() >= interval_seconds:
                self.last_run = current_time
                yield await self.produce()
            
            await asyncio.sleep(interval_seconds)

    async def close(self) -> None:
        """Clean up any resources."""
        pass


def register_data_source(cls):
    assert issubclass(
        cls, DataSource
    ), f"Class {cls.__name__} must be a subclass of DataSource to be registered as one."

    name = (
        cls.__name__
        if not cls.__name__.endswith("DataSource")
        else cls.__name__.removesuffix("DataSource")
    )
    __DATA_SOURCES[name] = cls
    register_config(cls, DataSourceConfig)
    return cls


__DATA_SOURCES = {}


def create_data_source(
    config: DataSourceConfig, alert_service: AlertService
) -> DataSource:
    """Create a data source based on the configuration.

    Args:
        config (DataSourceConfig): The configuration for the data source.
        alert_service (AlertService): The alert service to use.

    Raises:
        Exception: If the data source type is unknown.

    Returns:
        DataSource: The data source subclass instance.
    """
    c = config.config if hasattr(config, "config") else None

    src = __DATA_SOURCES.get(config.type)
    if src is not None:
        return src(alert_service, c)
    else:
        raise Exception(f"Unknown data source type: {config.type}")



from typing import Any
import psycopg2

from bs_monitoring.common.utils import ConfigField
from bs_monitoring.databases.base import Database, register_database


@register_database
class PostgresDatabase(Database):
    host = ConfigField(str)
    port = ConfigField(int)
    user = ConfigField(str)
    password = ConfigField(str)
    database = ConfigField(str)

    def __init__(self, config: Any) -> None:
        self.config = config
        self.conn = None

    def connect(self) -> None:
        self.conn = psycopg2.connect(
            host=self.host,
            port=self.port,
            user=self.user,
            password=self.password,
            database=self.database,
        )

    def close(self) -> None:
        if self.conn is not None:
            self.conn.close()

    def execute(self, query: str, params=None) -> None:
        cursor = self.conn.cursor()

        if params is not None:
            cursor.execute(query, params)
        else:
            cursor.execute(query)

        return cursor

    def commit(self) -> None:
        self.conn.commit()

    @property
    def autocommit(self) -> bool:
        return self.conn.autocommit

    @autocommit.setter
    def autocommit(self, value: bool) -> None:
        self.conn.autocommit = value



from abc import ABC, abstractmethod
from dataclasses import dataclass, asdict
from typing import Any

from bs_monitoring.alert_services.base import AlertService
from bs_monitoring.databases.base import DatabaseContext
from bs_monitoring.common.utils import register_config


@dataclass
class MonitorConfig:
    """Configuration for the monitor component

    Args:
        type (str): The type of the monitor.
        config (Union[DataSchemeMonitorConfig, None], optional): The configuration for the monitor. Defaults to None.
    """

    type: str
    db_name: str | None = None
    config: None = None


class Monitor(ABC):
    def __init__(
        self,
        alert_service: AlertService,
        db_name: str | None = None,
        config: Any = None,
    ) -> None:
        """Abstract class for monitors.

        Args:
            alert_service (AlertService): The alert service to use.
        """
        self.alert_service = alert_service
        self.connection = DatabaseContext.get_connection(db_name) if db_name else None

        if config:
            for k, v in asdict(config).items():
                setattr(self, k, v)

    @abstractmethod
    async def process(self, data: dict[str, list[dict[str, Any]]]) -> None:
        """Process the data from the data source.
        
        Args:
            data: Dictionary containing lists of events from different indices
        """
        pass


def register_monitor(cls: type[Monitor]) -> type[Monitor]:
    """Decorator to register a monitor.

    Args:
        cls (Monitor): Subclass of Monitor to register.

    Returns:
        Monitor: The registered Monitor subclass.
    """
    assert issubclass(
        cls, Monitor
    ), f"Class {cls.__name__} must be a subclass of Monitor to be registered as one."
    name = (
        cls.__name__
        if not cls.__name__.endswith("Monitor")
        else cls.__name__.removesuffix("Monitor")
    )

    __MONITORS[name] = cls
    register_config(cls, MonitorConfig)
    return cls


__MONITORS = {}


def create_monitors(
    configs: list[MonitorConfig], alert_service: AlertService
) -> list[Monitor]:
    """Method to create a monitor instances based on the configuration.

    Args:
        configs (List[MonitorConfig]): The configuration for the different monitors.
        alert_service (AlertService): The alert service to use.

    Returns: List[Monitor]: The list of monitor instances.
    """
    monitors = []
    for config in configs:
        c = getattr(config, "config", None)
        m = __MONITORS[config.type](alert_service, config.db_name, c)
        monitors.append(m)

    return monitors



import pkgutil
import importlib


def import_submodules(package_name: str) -> None:
    package = importlib.import_module(package_name)
    for _, name, _ in pkgutil.walk_packages(package.__path__):
        importlib.import_module(f"{package.__name__}.{name}")
